/*
Copyright (C) 2009  One Laptop Per Child Association Inc

Author: Sayamindu Dasgupta <sayamindu@laptop.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <Python.h>

#include <X11/Xlib.h>
#include <glib-object.h>
#include <libxklavier/xklavier.h>

XklEngine *engine;
XklConfigRegistry *config;
XklConfigRec *current_config;

static void
append_to_list_cb(XklConfigRegistry * config,
		  const XklConfigItem * item, gpointer data)
{
	PyObject *list;

	list = data;
	PyList_Append(list,
		      Py_BuildValue("(s,s)", item->name, item->description));
}

static void
append_to_list_withsubitem_cb(XklConfigRegistry * config,
			      const XklConfigItem * item,
			      const XklConfigItem * subitem, gpointer data)
{
	PyObject *list;

	list = data;
	if (subitem)
		PyList_Append(list,
			      Py_BuildValue("((s, s), (s,s))", item->name,
					    item->description, subitem->name,
					    subitem->description));
}

static PyObject *
pyxkbconfig_get_max_num_groups(PyObject *self, PyObject *args)
{
	gint num_groups;
	num_groups = xkl_engine_get_max_num_groups(engine);
	return Py_BuildValue("i", num_groups);
}

static PyObject *
pyxkbconfig_get_layouts(PyObject *self, PyObject *args)
{
	PyObject *layout_list;

	layout_list = PyList_New(0);
	xkl_config_registry_foreach_layout(config, append_to_list_cb,
					   layout_list);

	PyList_Sort(layout_list);
	return layout_list;
}

static PyObject *
pyxkbconfig_get_variants_for_layout(PyObject *self, PyObject *args)
{
	const gchar *layout;
	PyObject *variant_list;

	variant_list = PyList_New(0);
	if (!PyArg_ParseTuple(args, "s", &layout))
		return NULL;
	xkl_config_registry_foreach_layout_variant(config, layout,
						   append_to_list_cb,
						   variant_list);

	PyList_Sort(variant_list);
	return variant_list;
}

static PyObject *
pyxkbconfig_get_countries(PyObject *self, PyObject *args)
{
	PyObject *country_list;

	country_list = PyList_New(0);
	xkl_config_registry_foreach_country(config, append_to_list_cb,
					    country_list);

	PyList_Sort(country_list);
	return country_list;
}

static PyObject *
pyxkbconfig_get_variants_for_country(PyObject *self, PyObject *args)
{
	const gchar *country;
	PyObject *variant_list;

	variant_list = PyList_New(0);
	if (!PyArg_ParseTuple(args, "s", &country))
		return NULL;
	xkl_config_registry_foreach_country_variant(config, country,
						    append_to_list_withsubitem_cb,
						    variant_list);

	PyList_Sort(variant_list);
	return variant_list;
}

static PyObject *
pyxkbconfig_get_languages(PyObject *self, PyObject *args)
{
	PyObject *language_list;

	language_list = PyList_New(0);
	xkl_config_registry_foreach_language(config, append_to_list_cb,
					     language_list);

	PyList_Sort(language_list);
	return language_list;
}

static PyObject *
pyxkbconfig_get_variants_for_language(PyObject *self, PyObject *args)
{
	const gchar *language;
	PyObject *variant_list;

	variant_list = PyList_New(0);
	if (!PyArg_ParseTuple(args, "s", &language))
		return NULL;
	xkl_config_registry_foreach_language_variant(config, language,
						     append_to_list_withsubitem_cb,
						     variant_list);

	PyList_Sort(variant_list);
	return variant_list;
}


static PyObject *
pyxkbconfig_get_models(PyObject *self, PyObject *args)
{
	PyObject *model_list;

	model_list = PyList_New(0);
	xkl_config_registry_foreach_model(config, append_to_list_cb,
					  model_list);

	return model_list;
}

static PyObject *
pyxkbconfig_get_current_model(PyObject *self, PyObject *args)
{
	if (current_config->model != NULL)
		return Py_BuildValue("s", current_config->model);
	else
		Py_RETURN_NONE;
}


static PyObject *
pyxkbconfig_get_option_groups(PyObject *self, PyObject *args)
{
	PyObject *option_group_list;

	option_group_list = PyList_New(0);
	xkl_config_registry_foreach_option_group(config, append_to_list_cb,
						 option_group_list);

	PyList_Sort(option_group_list);
	return option_group_list;
}

static PyObject *
pyxkbconfig_get_options_for_option_group(PyObject *self, PyObject *args)
{
	const gchar *option_group;
	PyObject *option_list;

	option_list = PyList_New(0);
	if (!PyArg_ParseTuple(args, "s", &option_group))
		return NULL;
	xkl_config_registry_foreach_option(config, option_group,
					   append_to_list_cb, option_list);

	PyList_Sort(option_list);
	return option_list;
}


static PyObject *
pyxkbconfig_get_current_layouts(PyObject *self, PyObject *args)
{
	if (current_config->layouts == NULL)
		Py_RETURN_NONE;

	PyObject *layout_list;
	gint i;

	layout_list = PyList_New(0);
	for (i = 0; i < g_strv_length(current_config->layouts); i++)
		PyList_Append(layout_list,
			      Py_BuildValue("s", current_config->layouts[i]));

	return layout_list;
}

static PyObject *
pyxkbconfig_get_current_variants(PyObject *self, PyObject *args)
{
	if (current_config->variants == NULL)
		Py_RETURN_NONE;

	PyObject *variant_list;
	gint i;

	variant_list = PyList_New(0);
	for (i = 0; i < g_strv_length(current_config->variants); i++)
		PyList_Append(variant_list,
			      Py_BuildValue("s", current_config->variants[i]));

	return variant_list;
}

static PyObject *
pyxkbconfig_get_current_options(PyObject *self, PyObject *args)
{
	if (current_config->variants == NULL)
		Py_RETURN_NONE;

	PyObject *option_list;
	gint i;

	option_list = PyList_New(0);
	for (i = 0; i < g_strv_length(current_config->options); i++)
		PyList_Append(option_list,
			      Py_BuildValue("s", current_config->options[i]));

	return option_list;
}

static PyObject *
pyxkbconfig_set_layouts(PyObject *self, PyObject *args)
{
	PyObject *listObj;
	PyObject *strObj;
	gint num_layouts, i;
	gboolean ret_value;
	gchar *layout_arg;

	if (!PyArg_ParseTuple(args, "O!", &PyList_Type, &listObj))
		return NULL;

	num_layouts = PyList_Size(listObj);

	if (!num_layouts)
		return NULL;	/*FIXME: Raise exception */

	strObj = PyList_GetItem(listObj, 0);
	layout_arg = PyString_AsString(strObj);

	for (i = 1; i < num_layouts; i++) {
		strObj = PyList_GetItem(listObj, i);
		layout_arg = g_strjoin(",",
				       layout_arg, PyString_AsString(strObj),
				       NULL);
	}

	if (current_config->layouts != NULL)
		g_strfreev(current_config->layouts);
	current_config->layouts = g_new0(char *, 2);
	current_config->layouts[0] = g_strdup(layout_arg);
	g_free(layout_arg);

	/* We need to clean up the variant field of current_config 
	 * since it may cause an error if the variant already set 
	 * is not available for the new layout */
	if (current_config->variants != NULL)
		g_strfreev(current_config->variants);
	current_config->variants = g_new0(char *, 2);
	current_config->variants[0] = g_strdup("");

	ret_value = xkl_config_rec_activate(current_config, engine);

	return Py_BuildValue("b", ret_value);
}

static PyObject *
pyxkbconfig_set_variants(PyObject *self, PyObject *args)
{
	PyObject *listObj;
	PyObject *strObj;
	gint num_variants, i;
	gboolean ret_value;
	gchar *variant_arg;

	if (!PyArg_ParseTuple(args, "O!", &PyList_Type, &listObj))
		return NULL;

	num_variants = PyList_Size(listObj);

	if (!num_variants)
		return NULL;	/*FIXME: Raise exception */

	strObj = PyList_GetItem(listObj, 0);
	variant_arg = PyString_AsString(strObj);

	for (i = 1; i < num_variants; i++) {
		strObj = PyList_GetItem(listObj, i);
		variant_arg = g_strjoin(",",
					variant_arg, PyString_AsString(strObj),
					NULL);
	}

	if (current_config->variants != NULL)
		g_strfreev(current_config->variants);
	current_config->variants = g_new0(char *, 2);
	current_config->variants[0] = g_strdup(variant_arg);

	ret_value = xkl_config_rec_activate(current_config, engine);

	return Py_BuildValue("b", ret_value);
}

static PyObject *
pyxkbconfig_get_active_group_index(PyObject *self, PyObject *args)
{
	gint group_index;
	XklState *state;

	/* I'm not sure why the following line is not working in initpyxkb() */
	xkl_engine_start_listen(engine, XKLL_TRACK_KEYBOARD_STATE);
	state = xkl_engine_get_current_state(engine);
	group_index = state->group;
	xkl_engine_stop_listen(engine, XKLL_TRACK_KEYBOARD_STATE);

	return Py_BuildValue("i", group_index);
}


static PyObject *
pyxkbconfig_lock_next_group(PyObject *self, PyObject *args)
{
	gint next_group;

	next_group = xkl_engine_get_next_group(engine);
	xkl_engine_lock_group(engine, next_group);

	Py_RETURN_NONE;
}

static PyObject *
pyxkbconfig_lock_prev_group(PyObject *self, PyObject *args)
{
	gint prev_group;

	prev_group = xkl_engine_get_prev_group(engine);
	xkl_engine_lock_group(engine, prev_group);

	Py_RETURN_NONE;
}


/* Method table */
static PyMethodDef pyxkb_methods[] = {
	{"get_max_num_groups", pyxkbconfig_get_max_num_groups, METH_NOARGS,
	 "Returns maximum number of simultaneously supported groups (layouts)"},
	{"get_layouts", pyxkbconfig_get_layouts, METH_NOARGS,
	 "Returns list of available layouts"},
	{"get_variants_for_layout", pyxkbconfig_get_variants_for_layout,
	 METH_VARARGS, "Returns list of available variants for layout"},
	{"get_countries", pyxkbconfig_get_countries, METH_NOARGS,
	 "Returns list of available countries"},
	{"get_variants_for_country", pyxkbconfig_get_variants_for_country,
	 METH_VARARGS, "Returns list of available variants for country"},
	{"get_languages", pyxkbconfig_get_languages, METH_NOARGS,
	 "Returns list of available languages"},
	{"get_variants_for_language", pyxkbconfig_get_variants_for_language,
	 METH_VARARGS, "Returns list of available variants for language"},
	{"get_models", pyxkbconfig_get_models, METH_NOARGS,
	 "Returns list of available models"},
	{"get_option_groups", pyxkbconfig_get_option_groups, METH_NOARGS,
	 "Returns list of available option_groups"},
	{"get_options_for_option_group",
	 pyxkbconfig_get_options_for_option_group,
	 METH_VARARGS,
	 "Returns list of available options for option group"},
	{"get_current_model", pyxkbconfig_get_current_model, METH_NOARGS,
	 "Returns the currently enabled XKB model"},
	{"get_current_layouts", pyxkbconfig_get_current_layouts, METH_NOARGS,
	 "Returns list of currently enabled XKB layout(s)"},
	{"get_current_variants", pyxkbconfig_get_current_variants,
	 METH_NOARGS,
	 "Returns list of currently enabled XKB variant(s)"},
	{"get_current_options", pyxkbconfig_get_current_options,
	 METH_NOARGS,
	 "Returns list of currently enabled XKB options(s)"},
	{"set_layouts", pyxkbconfig_set_layouts, METH_VARARGS,
	 "Sets layout"},
	{"set_variants", pyxkbconfig_set_variants, METH_VARARGS,
	 "Sets variants (must be valid for the active layout(s))"},
	{"get_active_group_index", pyxkbconfig_get_active_group_index,
	 METH_NOARGS,
	 "Gets the index number of the currently active group"},
	{"lock_next_group", pyxkbconfig_lock_next_group, METH_NOARGS,
	 "Locks to the next group in the list of enabled groups"},
	{"lock_prev_group", pyxkbconfig_lock_prev_group, METH_NOARGS,
	 "Locks to the previous group in the list of enabled groups"},
	{NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC
initpyxkb(void)
{
	PyObject *m;
	Display *dpy;

	m = Py_InitModule("pyxkb", pyxkb_methods);

	g_type_init();

	dpy = XOpenDisplay(NULL);
	if (dpy == NULL) {
		PyErr_SetString(PyExc_RuntimeError, "Could not open display");
		return;
	}

	engine = xkl_engine_get_instance(dpy);

	config = xkl_config_registry_get_instance(engine);
	xkl_config_registry_load(config, 0);

	current_config = xkl_config_rec_new();
	xkl_config_rec_get_from_server(current_config, engine);
}

int
main(int argc, char **argv)
{
	Py_SetProgramName(argv[0]);
	Py_Initialize();
	initpyxkb();

	return 0;
}
